@@ -773,7 +773,7 @@ class CommandTree(Generic[ClientT]):
             for key in remove:
                 del mapping[key]
 
-    async def on_error(self, interaction: Interaction, error: AppCommandError) -> None:
+    async def on_error(self, interaction: Interaction, error: AppCommandError, /) -> None:
         """|coro|
 
         A callback that is called when any command raises an :exc:`AppCommandError`.
@@ -827,8 +827,7 @@ class CommandTree(Generic[ClientT]):
         if len(params) != 2:
             raise TypeError('error handler must have 2 parameters')
 
-        # Type checker doesn't like overriding methods like this
-        self.on_error = coro  # type: ignore
+        self.on_error = coro
         return coro
 
     def command(
@@ -842,7 +841,7 @@ class CommandTree(Generic[ClientT]):
         auto_locale_strings: bool = True,
         extras: Dict[Any, Any] = MISSING,
     ) -> Callable[[CommandCallback[Group, P, T]], Command[Group, P, T]]:
-        """Creates an application command directly under this tree.
+        """A decorator that creates an application command from a regular function directly under this tree.
 
         Parameters
         ------------
@@ -911,7 +910,7 @@ class CommandTree(Generic[ClientT]):
         auto_locale_strings: bool = True,
         extras: Dict[Any, Any] = MISSING,
     ) -> Callable[[ContextMenuCallback], ContextMenu]:
-        """Creates an application command context menu from a regular function directly under this tree.
+        """A decorator that creates an application command context menu from a regular function directly under this tree.
 
         This function must have a signature of :class:`~discord.Interaction` as its first parameter
         and taking either a :class:`~discord.Member`, :class:`~discord.User`, or :class:`~discord.Message`,
@@ -1078,9 +1077,10 @@ class CommandTree(Generic[ClientT]):
     async def _dispatch_error(self, interaction: Interaction, error: AppCommandError, /) -> None:
         command = interaction.command
         interaction.command_failed = True
-        if isinstance(command, Command):
-            await command._invoke_error_handlers(interaction, error)
-        else:
+        try:
+            if isinstance(command, Command):
+                await command._invoke_error_handlers(interaction, error)
+        finally:
             await self.on_error(interaction, error)
 
     def _from_interaction(self, interaction: Interaction) -> None:
