@@ -2435,7 +2435,7 @@ def dynamic_cooldown(
     """A decorator that adds a dynamic cooldown to a :class:`.Command`
 
     This differs from :func:`.cooldown` in that it takes a function that
-    accepts a single parameter of type :class:`.discord.Message` and must
+    accepts a single parameter of type :class:`.Context` and must
     return a :class:`~discord.app_commands.Cooldown` or ``None``.
     If ``None`` is returned then that cooldown is effectively bypassed.
 
