@@ -1,9 +1,7 @@
-# -*- coding: utf-8 -*-
-
 """
 The MIT License (MIT)
 
-Copyright (c) 2015-2020 Rapptz
+Copyright (c) 2015-present Rapptz
 
 Permission is hereby granted, free of charge, to any person obtaining a
 copy of this software and associated documentation files (the "Software"),
@@ -24,8 +22,30 @@ FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 DEALINGS IN THE SOFTWARE.
 """
 
-from . import utils
+from __future__ import annotations
+
 from .mixins import Hashable
+from .utils import snowflake_time, MISSING
+
+from typing import (
+    SupportsInt,
+    TYPE_CHECKING,
+    Type,
+    Union,
+)
+
+if TYPE_CHECKING:
+    import datetime
+    from . import abc
+
+    SupportsIntCast = Union[SupportsInt, str, bytes, bytearray]
+
+# fmt: off
+__all__ = (
+    'Object',
+)
+# fmt: on
+
 
 class Object(Hashable):
     """Represents a generic Discord object.
@@ -59,15 +79,39 @@ class Object(Hashable):
     -----------
     id: :class:`int`
         The ID of the object.
+    type: Type[:class:`abc.Snowflake`]
+        The discord.py model type of the object, if not specified, defaults to this class.
+
+        .. note::
+
+            In instances where there are multiple applicable types, use a shared base class.
+            for example, both :class:`Member` and :class:`User` are subclasses of :class:`abc.User`.
+
+        .. versionadded:: 2.0
     """
 
-    def __init__(self, id):
-        self.id = id
+    def __init__(self, id: SupportsIntCast, *, type: Type[abc.Snowflake] = MISSING):
+        try:
+            id = int(id)
+        except ValueError:
+            raise TypeError(f'id parameter must be convertible to int not {id.__class__!r}') from None
+        self.id: int = id
+        self.type: Type[abc.Snowflake] = type or self.__class__
 
-    def __repr__(self):
-        return '<Object id=%r>' % self.id
+    def __repr__(self) -> str:
+        return f'<Object id={self.id!r} type={self.type!r}>'
+
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, self.type):
+            return self.id == other.id
+        return NotImplemented
+
+    __hash__ = Hashable.__hash__
 
     @property
-    def created_at(self):
+    def created_at(self) -> datetime.datetime:
         """:class:`datetime.datetime`: Returns the snowflake's creation time in UTC."""
-        return utils.snowflake_time(self.id)
+        return snowflake_time(self.id)
+
+
+OLDEST_OBJECT = Object(id=0)
